  function [t,y] = adaptHeun(f,t0,T,y0,h0,atol,rtol)

% function adaptHeun
%
% Zweck: integriere die gewoehnliche Diff.-Gleichung erster Ordnung
%
%          dy/dt = f(t,y(t))
%
%        mit einem (zu?) einfachen adaptiven Heun Verfahren mit
%        Schrittweitenhalbierung-Fehlerschaetzer.
%
% Parameters: f    ... rechte Seite f(t,y(t)) der gew. Diff.-Gl.
%             t0,T ... Start- und End-Zeit
%             y0   ... Anfagswert
%             h0   ... Anfangs-Schrittweite
%             atol ... Absolute Toleranz
%             rtol ... Relative Toleranz
%
% Returns: t ... Zeiten
%          y ... approx. Loesung zu Zeiten t
%
% Notes: Keine.
%

% Parameter
  p = 2;         % KO des Heun Verfahren
  hmin = 10*eps; % kleinste zugelassene Schrittweite
  fac = 0.9;     % Sicherheits-Faktor
  facmin = 0.5;  % kleinste zugelassene Schrittweiten-Verkleinerung
  facmax = 1.5;  % groesste zugelassene Schrittweiten-Vergroesserung

% Initialisiere Zeit- und Loesungs-Vektor
  t = t0;
  y = y0;

% Mache Zeitschritte solange Endzeit nicht erreicht
  j  = 0;    % Initialisiere Schritt-Zaehler
  Fails = 0; % Initialisiere gescheiterte Schritt-Zaehler
  hj = h0;   % Initialisiere Schrittweite
  epsilonjp = atol + norm(y0)*rtol; % Initialisiere Fehlerschaetzer
  while (t(end) < T)

    % Zwischenspeicherung der Daten zum j-ten Schritt
    tj = t(end);
    yj = y(:,end);

    % Schaetze Schrittweite hj
    tolj = atol + norm(yj)*rtol;
    hj = hj*min(facmax,max(facmin,fac*(tolj/epsilonjp)^(1./(p + 1.))));

    % Fange zu kleine Schrittweiten ab
    if ( hj < hmin )
      break
    end

    % Korrigiere Schrittweite hj so, dass die Endzeit T genau erreicht wird
    hj = min(T - tj,hj);

    % Heun Verfahren mit vollem Schritt h   --> y_{j+1} == yjp
    k1 = f(tj,yj);
    k2 = f(tj + hj,yj + hj*k1);
    yjp = yj +  0.5*hj*(k1 + k2);         % y_{j+1}
    % Heun Verfahren mit halbem Schritt h/2 --> yhut_{j+1} == yhutjp
    % Erster halb-Schritt
    % k1 = f(tj,yj); % koennen wir von oben uebernehmen!
                     % (damit sparen wir uns eine f Auswerung :-)
    k2 = f(tj + 0.5*hj,yj + 0.5*hj*k1);
    yhutjph = yj +  0.25*hj*(k1 + k2);    % yhut_{j+1/2} = yhutjph
    % Zweiter halb-Schritt
    k1 = f(tj + 0.5*hj,yhutjph);
    k2 = f(tj + hj,yhutjph + 0.5*hj*k1);
    yhutjp = yhutjph + 0.25*hj*(k1 + k2); % yhut_{j+1} = yhutjp

    % Fehlerschaetzer (fuer yhut_{j+1} = yhutjp, d.h. das genauere Verfahren!)
    epsilonjp = norm(yhutjp - yjp)/(2^p - 1.);

    % Kontrolliere Toleranz-Kriterium
    if ( epsilonjp < tolj ) % akzeptiere Zeitschritt!
      j = j + 1;
      t = [t,tj+hj];
      y = [y,yhutjp];
    else % verwerfe Zeitschritt und halbiere Schrittweite
      Fails = Fails + 1;
    end

  end

  if ( abs(t(end) - T) < eps ) % Integration erfolgreich!
    fprintf('Info: Anzahl akzeptierte Int.-Schritte %i \n',j);
    fprintf('Info: Anzahl verworfene  Int.-Schritte %i \n',Fails);
    fprintf('Info: Verhaeltnis %e \n',Fails/j);
  else % Integration geschweitert!
    fprintf('Error: Gewuenschte Endzeit nicht erreicht!!! \n');
    fprintf('Error: Letzte Schrittweite hj = %e \n',hj);
  end

